from . import errors

AUTO_UPDATE_UNCHANGED=object()

class Property(object):
    data_type = None

    def __init__(self, verbose_name=None, name=None,
                 default=None, required=False,
                 validator=None, choices=None, indexed=True):
        self.verbose_name = verbose_name
        self.name = name
        self.default = default
        self.required = required
        self.validator = validator
        self.choices = choices
        self.indexed = indexed

    def __get__(self, inst, cls):
        if inst is None: return self
        try:
            return inst[self._attr_name()]
        except KeyError:
            return None

    def __set__(self, inst, value):
        value = self.validate(value)
        inst[self._attr_name()] = value
        
    def default_value(self):
        return self.default

    def validate(self, value):
        if self.empty(value):
          if self.required:
            raise errors.BadValueError('Property %s is required' % self.name)
        else:
          if self.choices:
            match = False
            for choice in self.choices:
              if choice == value:
                match = True
            if not match:
              raise errors.BadValueError('Property %s is %r; must be one of %r' %
                                         (self.name, value, self.choices))
        if self.validator is not None:
          self.validator(value)
        return value

    def empty(self, value):
        return not value

    def get_value_for_datastore(self, inst):
        return self.__get__(inst, inst.__class__)

    def make_value_from_datastore(self, value):
        return value

    def make_value_from_datastore_index_value(self, value):
        return value

    @staticmethod
    def _require_parameter(kwds, parameter, value):
        if parameter in kwds and kwds[parameter] != value:
            raise errors.ConfigurationError('%s must be %s.' % (parameter, value))
        kwds[parameter] = value

    
class UnindexedProperty(Property):

    def __init__(self, *args, **kwargs):
        self._require_parameter(kwargs, 'indexed', False)
        super(UnindexedProperty, self).__init__(self, *args, **kwargs)

    def validate(self, value):
        if value is not None and not isinstance(value, self.data_type):
            try:
                value = self.data_type(value)
            except TypeError, err:
                raise errors.BadValueError(
                    'Property %s must be convertible '
                    'to a %s instance (%s)' %
                    (self.name, self.data_type.__name__, err))
        value = super(UnindexedProperty, self).validate(value)
        if value is not None and not isinstance(value, self.data_type):
            raise errors.BadValueError(
                'Property %s must be a %s instance' %
                (self.name, self.data_type.__name__))
        return value

class TextProperty(UnindexedProperty):
    data_type = unicode

class StringProperty(Property):
    data_type = basestring
    MAX_LENGTH = 500

    def __init__(self, *args, **kwargs):
        self.multiline = kwargs.pop('multiline', False)
        super(StringProperty, self).__init__(*args, **kwargs)

    def validate(self, value):
        if value is not None and not isinstance(value, basestring):
            raise errors.BadValueError(
                'Property %s must be a str or unicode instance, not a %s'
                % (self.name, type(value).__name__))
        if not self.multiline and value and value.find('\n') != -1:
            raise errors.BadValueError('Property %s is not multi-line' % self.name)
        if value is not None and len(value) > self.MAX_LENGTH:
            raise errors.BadValueError(
                'Property %s is %d characters long; it must be %d or less.'
                % (self.name, len(value), self.MAX_LENGTH))
        return value

class _CoercingProperty(Property):

  def validate(self, value):
    value = super(_CoercingProperty, self).validate(value)
    if value is not None and not isinstance(value, self.data_type):
      value = self.data_type(value)
    return value






    
